import glob, re
import json, os, dataclasses
from datetime import datetime
from typing import Dict, List, Optional
import configparser
import stackprinter

from qubx.core.basics import Instrument, FuturesInfo, TransactionCostsCalculator, ZERO_COSTS
from qubx.utils.marketdata.binance import get_binance_symbol_info_for_type
from qubx import logger
from qubx.utils.misc import makedirs, get_local_qubx_folder

_DEF_INSTRUMENTS_FOLDER = "instruments"
_DEF_FEES_FOLDER = "fees"


class _InstrumentEncoder(json.JSONEncoder):
    def default(self, obj):
        if dataclasses.is_dataclass(obj):
            return {k: v for k, v in dataclasses.asdict(obj).items() if not k.startswith("_")}
        if isinstance(obj, (datetime)):
            return obj.isoformat()
        return super().default(obj)


class _InstrumentDecoder(json.JSONDecoder):
    def _preprocess(d, ks):
        fi = d.get(ks)
        _preproc = lambda x: x[: x.find(".")] if x.endswith("Z") else x
        if fi:
            fi["delivery_date"] = datetime.strptime(
                _preproc(fi.get("delivery_date", "5000-01-01T00:00:00")), "%Y-%m-%dT%H:%M:%S"
            )
            fi["onboard_date"] = datetime.strptime(
                _preproc(fi.get("onboard_date", "1970-01-01T00:00:00")), "%Y-%m-%dT%H:%M:%S"
            )
        return d | {ks: FuturesInfo(**fi) if fi else None}

    def decode(self, json_string):
        obj = super(_InstrumentDecoder, self).decode(json_string)
        if isinstance(obj, dict):
            return Instrument(**_InstrumentDecoder._preprocess(obj, "futures_info"))
        elif isinstance(obj, list):
            return [Instrument(**_InstrumentDecoder._preprocess(o, "futures_info")) for o in obj]
        return obj


class InstrumentsLookup:
    _lookup: Dict[str, Instrument]
    _path: str

    def __init__(self, path: str = makedirs(get_local_qubx_folder(), _DEF_INSTRUMENTS_FOLDER)) -> None:
        self._path = path
        if not self.load():
            self.refresh()
        self.load()

    def load(self) -> bool:
        self._lookup = {}
        data_exists = False
        for fs in glob.glob(self._path + "/*.json"):
            try:
                with open(fs, "r") as f:
                    instrs = json.load(f, cls=_InstrumentDecoder)
                    for i in instrs:
                        self._lookup[f"{i.exchange}:{i.symbol}"] = i
                    data_exists = True
            except Exception as ex:
                stackprinter.show_current_exception()
                logger.warning(ex)

        return data_exists

    def find(self, exchange: str, base: str, quote: str, margin: Optional[str] = None) -> Optional[Instrument]:
        for i in self._lookup.values():
            if i.exchange == exchange and (
                (i.base == base and i.quote == quote) or (i.base == quote and i.quote == base)
            ):
                if margin is not None and i.margin_symbol is not None:
                    if i.margin_symbol == margin:
                        return i
                else:
                    return i
        return None

    def find_symbol(self, exchange: str, symbol: str) -> Optional[Instrument]:
        for i in self._lookup.values():
            if (i.exchange == exchange) and (i.symbol == symbol):
                return i
        return None

    def find_instruments(self, exchange: str, quote: str | None = None) -> list[Instrument]:
        return [i for i in self._lookup.values() if i.exchange == exchange and (quote is None or i.quote == quote)]

    def _save_to_json(self, path, instruments: List[Instrument]):
        with open(path, "w") as f:
            json.dump(instruments, f, cls=_InstrumentEncoder)
        logger.info(f"Saved {len(instruments)} to {path}")

    def find_aux_instrument_for(self, instrument: Instrument, base_currency: str) -> Optional[Instrument]:
        """
        Tries to find aux instrument (for conversions to funded currency)
        for example:
            ETHBTC -> BTCUSDT for base_currency USDT
            EURGBP -> GBPUSD for base_currency USD
            ...
        """
        base_currency = base_currency.upper()
        if instrument.quote != base_currency and instrument._aux_instrument is None:
            return self.find(instrument.exchange, instrument.quote, base_currency)
        return instrument._aux_instrument

    def __getitem__(self, spath: str) -> List[Instrument]:
        res = []
        c = re.compile(spath)
        for k, v in self._lookup.items():
            if re.match(c, k):
                res.append(v)
        return res

    def refresh(self):
        for mn in dir(self):
            if mn.startswith("_update_"):
                getattr(self, mn)(self._path)

    def _update_kraken(self, path: str):
        import ccxt as cx

        kf = cx.krakenfutures()
        ks = cx.kraken()
        f_mkts = kf.load_markets()
        s_mkts = ks.load_markets()

        # - process futures
        f_instruments = []
        for _, v in f_mkts.items():
            info = v["info"]
            # - we skip index as it's not traded
            if v["index"]:
                continue
            maint_margin = 0
            required_margin = 0
            if "marginLevels" in info:
                margins = info["marginLevels"][0]
                maint_margin = float(margins["maintenanceMargin"])
                required_margin = float(margins["initialMargin"])
            f_instruments.append(
                Instrument(
                    v["symbol"],
                    v["type"],
                    "KRAKEN.F",
                    v["base"],
                    v["quote"],
                    v["settle"],
                    min_tick=float(v["precision"]["price"]),
                    min_size_step=float(v["precision"]["amount"]),
                    min_size=v["precision"]["amount"],
                    futures_info=FuturesInfo(
                        contract_type=info["type"],
                        contract_size=float(info["contractSize"]),
                        onboard_date=info["openingDate"],
                        delivery_date=v["expiryDatetime"] if "expiryDatetime" in info else "2100-01-01T00:00:00",
                        maint_margin=maint_margin,
                        required_margin=required_margin,
                    ),
                )
            )
        # - drop to file
        self._save_to_json(os.path.join(path, "kraken.f.json"), f_instruments)

        # - process spots
        s_instruments = []
        for _, v in s_mkts.items():
            info = v["info"]
            s_instruments.append(
                Instrument(
                    v["symbol"],
                    v["type"],
                    "KRAKEN",
                    v["base"],
                    v["quote"],
                    v["settle"],
                    min_tick=float(info["tick_size"]),
                    min_size_step=float(v["precision"]["price"]),
                    min_size=float(v["precision"]["amount"]),
                )
            )
        # - drop to file
        self._save_to_json(os.path.join(path, "kraken.json"), s_instruments)

    def _update_dukas(self, path: str):
        instruments = [
            Instrument("EURUSD", "FX", "DUKAS", "EUR", "USD", "USD", 0.00001, 1, 1000),
            Instrument("GBPUSD", "FX", "DUKAS", "GBP", "USD", "USD", 0.00001, 1, 1000),
            Instrument("USDJPY", "FX", "DUKAS", "USD", "JPY", "USD", 0.001, 1, 1000),
            Instrument("USDCAD", "FX", "DUKAS", "USD", "CAD", "USD", 0.00001, 1, 1000),
            Instrument("AUDUSD", "FX", "DUKAS", "AUD", "USD", "USD", 0.00001, 1, 1000),
            Instrument("USDPLN", "FX", "DUKAS", "USD", "PLN", "USD", 0.00001, 1, 1000),
            Instrument("EURGBP", "FX", "DUKAS", "EUR", "GBP", "USD", 0.00001, 1, 1000),
            # TODO: addd all or find how to get it from site
        ]
        self._save_to_json(os.path.join(path, "dukas.json"), instruments)

    def _update_binance(self, path: str):
        infos = get_binance_symbol_info_for_type(["UM", "CM", "SPOT"])
        for exchange, info in infos.items():
            instruments = []
            for s in info["symbols"]:
                tick_size, size_step = None, None
                for i in s["filters"]:
                    if i["filterType"] == "PRICE_FILTER":
                        tick_size = float(i["tickSize"])
                    if i["filterType"] == "LOT_SIZE":
                        size_step = float(i["stepSize"])

                fut_info = None
                if "contractType" in s:
                    fut_info = FuturesInfo(
                        s.get("contractType", "UNKNOWN"),
                        datetime.fromtimestamp(s.get("deliveryDate", 0) / 1000.0),
                        datetime.fromtimestamp(s.get("onboardDate", 0) / 1000.0),
                        float(s.get("contractSize", 1)),
                        float(s.get("maintMarginPercent", 0)),
                        float(s.get("requiredMarginPercent", 0)),
                        float(s.get("liquidationFee", 0)),
                    )

                instruments.append(
                    Instrument(
                        s["symbol"],
                        "CRYPTO",
                        exchange.upper(),
                        s["baseAsset"],
                        s["quoteAsset"],
                        s.get("marginAsset", None),
                        tick_size,
                        size_step,
                        min_size=size_step,  # TODO: not sure about minimal position for Binance
                        futures_info=fut_info,
                    )
                )
            # - store to file
            self._save_to_json(os.path.join(path, f"{exchange}.json"), instruments)


# - TODO: need to find better way to extract actual data !!
_DEFAULT_FEES = """
[binance]
# SPOT (maker, taker)
vip0_usdt = 0.1000,0.1000
vip1_usdt = 0.0900,0.1000
vip2_usdt = 0.0800,0.1000
vip3_usdt = 0.0420,0.0600
vip4_usdt = 0.0420,0.0540
vip5_usdt = 0.0360,0.0480
vip6_usdt = 0.0300,0.0420
vip7_usdt = 0.0240,0.0360
vip8_usdt = 0.0180,0.0300
vip9_usdt = 0.0120,0.0240

# SPOT (maker, taker)
vip0_bnb = 0.0750,0.0750
vip1_bnb = 0.0675,0.0750
vip2_bnb = 0.0600,0.0750
vip3_bnb = 0.0315,0.0450
vip4_bnb = 0.0315,0.0405
vip5_bnb = 0.0270,0.0360
vip6_bnb = 0.0225,0.0315
vip7_bnb = 0.0180,0.0270
vip8_bnb = 0.0135,0.0225
vip9_bnb = 0.0090,0.0180

# UM futures (maker, taker)
[binance.um]
vip0_usdt = 0.0200,0.0500
vip1_usdt = 0.0160,0.0400
vip2_usdt = 0.0140,0.0350
vip3_usdt = 0.0120,0.0320
vip4_usdt = 0.0100,0.0300
vip5_usdt = 0.0080,0.0270
vip6_usdt = 0.0060,0.0250
vip7_usdt = 0.0040,0.0220
vip8_usdt = 0.0020,0.0200
vip9_usdt = 0.0000,0.0170

# CM futures (maker, taker)
[binance.cm]
vip0 = 0.0200,0.0500
vip1 = 0.0160,0.0400
vip2 = 0.0140,0.0350
vip3 = 0.0120,0.0320
vip4 = 0.0100,0.0300
vip5 = 0.0080,0.0270
vip6 = 0.0060,0.0250
vip7 = 0.0040,0.0220
vip8 = 0.0020,0.0200
vip9 = 0.0000,0.0170

[bitmex]
tierb_xbt=0.02,0.075
tierb_usdt=-0.015,0.075
tieri_xbt=0.01,0.05
tieri_usdt=-0.015,0.05
tiert_xbt=0.0,0.04
tiert_usdt=-0.015,0.04
tierm_xbt=0.0,0.035
tierm_usdt=-0.015,0.035
tiere_xbt=0.0,0.03
tiere_usdt=-0.015,0.03
tierx_xbt=0.0,0.025
tierx_usdt=-0.015,0.025
tierd_xbt=-0.003,0.024
tierd_usdt=-0.015,0.024
tierw_xbt=-0.005,0.023
tierw_usdt=-0.015,0.023
tierk_xbt=-0.008,0.022
tierk_usdt=-0.015,0.022
tiers_xbt=-0.01,0.0175
tiers_usdt=-0.015,0.02

[dukas]
regular=0.0035,0.0035
premium=0.0017,0.0017

[kraken]
K0=0.25,0.40
K10=0.20,0.35
K50=0.14,0.24
K100=0.12,0.22
K250=0.10,0.20
K500=0.08,0.18
M1=0.06,0.16
M2.5=0.04,0.14
M5=0.02,0.12
M10=0.0,0.10

[kraken.f]
K0=0.0200,0.0500
K100=0.0150,0.0400
M1=0.0125,0.0300
M5=0.0100,0.0250
M10=0.0075,0.0200
M20=0.0050,0.0150
M50=0.0025,0.0125
M100=0.0000,0.0100
"""


class FeesLookup:
    """
    Fees lookup
    """

    _lookup: Dict[str, TransactionCostsCalculator]
    _path: str

    def __init__(self, path: str = makedirs(get_local_qubx_folder(), _DEF_FEES_FOLDER)) -> None:
        self._path = path
        if not self.load():
            self.refresh()
        self.load()

    def load(self) -> bool:
        self._lookup = {}
        data_exists = False
        parser = configparser.ConfigParser()
        # - load all avaliable configs
        for fs in glob.glob(self._path + "/*.ini"):
            parser.read(fs)
            data_exists = True

        for exch in parser.sections():
            for spec, info in parser[exch].items():
                try:
                    maker, taker = info.split(",")
                    self._lookup[f"{exch}_{spec}"] = (float(maker), float(taker))
                except:
                    logger.warning(f'Wrong spec format for {exch}: "{info}". Should be spec=maker,taker')

        return data_exists

    def __getitem__(self, spath: str) -> List[Instrument]:
        res = []
        c = re.compile(spath)
        for k, v in self._lookup.items():
            if re.match(c, k):
                res.append((k, v))
        return res

    def refresh(self):
        with open(os.path.join(self._path, "default.ini"), "w") as f:
            f.write(_DEFAULT_FEES)

    def find(self, exchange: str, spec: str | None) -> Optional[TransactionCostsCalculator]:
        if spec is None:
            return ZERO_COSTS
        key = f"{exchange}_{spec}"
        vals = self._lookup.get(key)
        return TransactionCostsCalculator(key, *self._lookup.get(key)) if vals is not None else None

    def __repr__(self) -> str:
        s = "Name:\t\t\t(maker, taker)\n"
        for k, v in self._lookup.items():
            s += f"{k.ljust(25)}: {v}\n"
        return s


@dataclasses.dataclass(frozen=True)
class GlobalLookup:
    instruments: InstrumentsLookup
    fees: FeesLookup

    def find_fees(self, exchange: str, spec: str | None) -> Optional[TransactionCostsCalculator]:
        return self.fees.find(exchange, spec)

    def find_aux_instrument_for(self, instrument: Instrument, base_currency: str) -> Optional[Instrument]:
        return self.instruments.find_aux_instrument_for(instrument, base_currency)

    def find_instrument(self, exchange: str, base: str, quote: str) -> Optional[Instrument]:
        return self.instruments.find(exchange, base, quote)

    def find_instruments(self, exchange: str, quote: str | None = None) -> list[Instrument]:
        return self.instruments.find_instruments(exchange, quote)

    def find_symbol(self, exchange: str, symbol: str) -> Optional[Instrument]:
        return self.instruments.find_symbol(exchange, symbol)



from typing import Any, Dict, List, Optional, Tuple

from collections import defaultdict
import stackprinter
import traceback

import ccxt
from ccxt.base.exchange import Exchange, ExchangeError

import numpy as np
import pandas as pd

from qubx import logger, lookup
from qubx.core.account import AccountProcessor
from qubx.core.basics import Instrument, Position, Order, TransactionCostsCalculator, dt_64, Deal, CtrlChannel
from qubx.core.interfaces import IBrokerServiceProvider, ITradingServiceProvider
from qubx.core.series import TimeSeries, Bar, Trade, Quote
from qubx.connectors.ccxt.ccxt_utils import (
    EXCHANGE_ALIASES,
    ccxt_convert_order_info,
    ccxt_convert_deal_info,
    ccxt_extract_deals_from_exec,
    ccxt_restore_position_from_deals,
)
from qubx.utils.ntp import time_now


ORDERS_HISTORY_LOOKBACK_DAYS = 30


class CCXTTradingConnector(ITradingServiceProvider):
    """
    Synchronous instance of trading API
    """

    sync: Exchange

    _fees_calculator: TransactionCostsCalculator
    _positions: Dict[Instrument, Position]

    def __init__(
        self,
        exchange_id: str,
        account_id: str,
        commissions: str | None = None,
        **exchange_auth,
    ):
        exchange_id = exchange_id.lower()
        exch = EXCHANGE_ALIASES.get(exchange_id, exchange_id)
        if exch not in ccxt.exchanges:
            raise ValueError(f"Exchange {exchange_id} -> {exch} is not supported by CCXT!")

        self.exchange_id = exchange_id
        self.account_id = account_id
        self.commissions = commissions

        # - sync exchange
        self.sync: Exchange = getattr(ccxt, exch.lower())(exchange_auth)

        logger.info(f"{exch.upper()} loading ...")
        self.sync.load_markets()

    def set_account(self, acc: AccountProcessor):
        super().set_account(acc)
        # TODO: move sync account info call to base class, could be useful also for IB
        self._sync_account_info(self.commissions)
        self._positions = self.acc._positions
        # - show reserves info
        for s, v in self.acc.reserved.items():
            logger.info(f" > {v} of {s} is reserved from trading")

    def _sync_account_info(self, default_commissions: str | None):
        logger.info(f"Loading account data for {self.get_name()}")
        self._balance = self.sync.fetch_balance()
        _info = self._balance.get("info")

        # - check what we have on balance: TODO test on futures account
        for k, vol in self._balance["total"].items():  # type: ignore
            if vol != 0.0:  # - get all non zero balances
                self.acc.update_balance(k, vol, vol - self._balance["free"].get(k, 0))

        # - try to get account's commissions calculator or set default one
        if _info:
            _fees = _info.get("commissionRates")
            if _fees:
                self._fees_calculator = TransactionCostsCalculator(
                    "account", 100 * float(_fees["maker"]), 100 * float(_fees["taker"])
                )

        if self._fees_calculator is None:
            if default_commissions:
                self._fees_calculator = lookup.fees.find(self.exchange_id.lower(), default_commissions)
            else:
                raise ValueError("Can't get commissions level from account, but default commissions is not defined !")

    def _get_open_orders_from_exchange(self, instrument: Instrument, days_before: int = 60) -> Dict[str, Order]:
        """
        We need only open orders to restore list of active ones in connector
        method returns open orders sorted by creation time in ascending order
        """
        t_orders_start_ms = (self.time() - days_before * pd.Timedelta("1d")).asm8.item() // 1000000
        orders_data = self.sync.fetch_open_orders(instrument.symbol, since=t_orders_start_ms)
        orders: Dict[str, Order] = {}
        for o in orders_data:
            order = ccxt_convert_order_info(instrument, o)
            orders[order.id] = order
        if orders:
            logger.info(f"{instrument.symbol} - loaded {len(orders)} open orders")
        return dict(sorted(orders.items(), key=lambda x: x[1].time, reverse=False))

    def _get_deals_from_exchange(self, symbol: str, days_before: int = 60) -> List[Deal]:
        """
        Load trades for given symbol
        method returns account's trades sorted by creation time in reversed order (latest - first)
        """
        t_orders_start_ms = (self.time() - days_before * pd.Timedelta("1d")).asm8.item() // 1000000
        deals_data = self.sync.fetch_my_trades(symbol, since=t_orders_start_ms)
        deals: List[Deal] = [ccxt_convert_deal_info(o) for o in deals_data]  # type: ignore
        if deals:
            return list(sorted(deals, key=lambda x: x.time, reverse=False))
        return list()

    def _sync_position_and_orders(self, position: Position) -> Position:
        asset = position.instrument.base
        symbol = position.instrument.symbol
        total_amnts = self._balance["total"]
        vol_from_exch = total_amnts.get(asset, total_amnts.get(symbol, 0))

        # - get orders from exchange
        orders = self._get_open_orders_from_exchange(position.instrument, ORDERS_HISTORY_LOOKBACK_DAYS)
        self.acc.add_active_orders(orders)

        # - get deals from exchange if position is not zero
        if vol_from_exch != 0:
            deals = self._get_deals_from_exchange(symbol)

            # - actualize position
            position = ccxt_restore_position_from_deals(
                position, vol_from_exch, deals, reserved_amount=self.acc.get_reserved(position.instrument)
            )

        return position

    def get_position(self, instrument: Instrument) -> Position:
        if instrument not in self._positions:
            position = Position(instrument)
            position = self._sync_position_and_orders(position)
            self.acc.attach_positions(position)
        return self._positions[instrument]

    def send_order(
        self,
        instrument: Instrument,
        order_side: str,
        order_type: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
    ) -> Order:
        params = {}

        if order_type == "limit":
            params["timeInForce"] = time_in_force.upper()
            if price is None:
                raise ValueError("Price must be specified for limit order")

        if client_id:
            params["newClientOrderId"] = client_id

        r: Dict[str, Any] | None = None
        try:
            r = self.sync.create_order(instrument.symbol, order_type, order_side, amount, price, params=params)  # type: ignore
        except ccxt.BadRequest as exc:
            logger.error(
                f"(CCXTSyncTradingConnector::send_order) BAD REQUEST for {order_side} {amount} {order_type} for {symbol} : {exc}"
            )
            raise exc
        except Exception as err:
            logger.error(
                f"(CCXTSyncTradingConnector::send_order) {order_side} {amount} {order_type} for {instrument.symbol} exception : {err}"
            )
            logger.error(traceback.format_exc())
            raise err

        if r is None:
            logger.error(f"(CCXTSyncTradingConnector::send_order) No response from exchange")
            raise ExchangeError("(CCXTSyncTradingConnector::send_order) No response from exchange")

        order = ccxt_convert_order_info(instrument, r)
        logger.info(f"(CCXTSyncTradingConnector) New order {order}")
        return order

    def cancel_order(self, order_id: str) -> Order | None:
        order = None
        if order_id in self.acc._active_orders:
            order = self.acc._active_orders[order_id]
            try:
                logger.info(f"Canceling order {order_id} ...")
                # r = self._task_s(self.exchange.cancel_order(order_id, symbol=order.instrument.symbol))
                r = self.sync.cancel_order(order_id, symbol=order.instrument.symbol)
            except Exception as err:
                logger.error(f"(CCXTSyncTradingConnector) canceling [{order}] exception : {err}")
                logger.error(traceback.format_exc())
                raise err
        return order

    def time(self) -> dt_64:
        """
        Returns current time as dt64
        """
        # return np.datetime64(self.sync.microseconds() * 1000, "ns")
        return time_now()

    def get_name(self) -> str:
        return self.sync.name  # type: ignore

    def get_account_id(self) -> str:
        return self.account_id

    def process_execution_report(self, instrument: Instrument, report: dict[str, Any]) -> Tuple[Order, List[Deal]]:
        order = ccxt_convert_order_info(instrument, report)
        deals = ccxt_extract_deals_from_exec(report)
        self._fill_missing_fee_info(instrument, deals)
        self.acc.process_deals(instrument, deals)
        self.acc.process_order(order)
        return order, deals

    def get_orders(self, instrument: Instrument | None = None) -> list[Order]:
        return self.acc.get_orders(instrument)

    def get_base_currency(self) -> str:
        return self.acc.base_currency

    def _fill_missing_fee_info(self, instrument: Instrument, deals: List[Deal]) -> None:
        for d in deals:
            if d.fee_amount is None:
                d.fee_amount = self._fees_calculator.get_execution_fees(
                    instrument=instrument, exec_price=d.price, amount=d.amount, crossed_market=d.aggressive
                )
                # this is only true for linear contracts
                d.fee_currency = instrument.quote



import pandas as pd
from qubx import lookup, logger
from qubx.core.account import AccountProcessor
from qubx.core.helpers import CachedMarketDataHolder
from qubx.core.loggers import StrategyLogging
from qubx.core.basics import TargetPosition, Instrument, Position
from qubx.core.loggers import StrategyLogging
from qubx.core.interfaces import (
    IBrokerServiceProvider,
    IPositionGathering,
    IStrategy,
    ITradingServiceProvider,
    IStrategyContext,
    IUniverseManager,
    ISubscriptionManager,
    ITradingManager,
    ITimeProvider,
)


class UniverseManager(IUniverseManager):
    __context: IStrategyContext
    __strategy: IStrategy
    __broker: IBrokerServiceProvider
    __trading_service: ITradingServiceProvider
    __cache: CachedMarketDataHolder
    __logging: StrategyLogging
    __subscription_manager: ISubscriptionManager
    __trading_manager: ITradingManager
    __time_provider: ITimeProvider
    __positions: dict[Instrument, Position]
    __position_gathering: IPositionGathering

    def __init__(
        self,
        context: IStrategyContext,
        strategy: IStrategy,
        broker: IBrokerServiceProvider,
        trading_service: ITradingServiceProvider,
        cache: CachedMarketDataHolder,
        logging: StrategyLogging,
        subscription_manager: ISubscriptionManager,
        trading_manager: ITradingManager,
        time_provider: ITimeProvider,
        account_processor: AccountProcessor,
        position_gathering: IPositionGathering,
    ):
        self.__context = context
        self.__strategy = strategy
        self.__broker = broker
        self.__trading_service = trading_service
        self.__cache = cache
        self.__logging = logging
        self.__subscription_manager = subscription_manager
        self.__trading_manager = trading_manager
        self.__time_provider = time_provider
        self.__positions = account_processor.positions
        self.__position_gathering = position_gathering
        self._instruments = []

    def set_universe(self, instruments: list[Instrument], skip_callback: bool = False) -> None:
        new_set = set(instruments)
        prev_set = set(self._instruments)
        rm_instr = list(prev_set - new_set)
        add_instr = list(new_set - prev_set)

        self.__add_instruments(add_instr)
        self.__remove_instruments(rm_instr)

        if not skip_callback and (add_instr or rm_instr):
            self.__strategy.on_universe_change(self.__context, add_instr, rm_instr)

        self.__broker.commit()  # apply pending changes

        # set new instruments
        self._instruments.clear()
        self._instruments.extend(instruments)

    @property
    def instruments(self) -> list[Instrument]:
        return self._instruments

    def __remove_instruments(self, instruments: list[Instrument]) -> None:
        """
        Remove symbols from universe. Steps:
        - close all open positions
        - unsubscribe from market data
        - remove from data cache

        We are still keeping the symbols in the positions dictionary.
        """
        # - close all open positions
        exit_targets = [
            TargetPosition.zero(self.__context, instr.signal(0, group="Universe", comment="Universe change"))
            for instr in instruments
            if instr.symbol in self.__positions and abs(self.__positions[instr.symbol].quantity) > instr.min_size
        ]
        self.__position_gathering.alter_positions(self.__context, exit_targets)

        # - if still open positions close them manually
        for instr in instruments:
            pos = self.__positions.get(instr)
            if pos and abs(pos.quantity) > instr.min_size:
                self.__trading_manager.trade(instr, -pos.quantity)

        # - unsubscribe from market data
        for instr in instruments:
            self.__subscription_manager.unsubscribe(instr)

        # - remove from data cache
        for instr in instruments:
            self.__cache.remove(instr)

    def __add_instruments(self, instruments: list[Instrument]) -> None:
        # - create positions for instruments
        self._create_and_update_positions(instruments)

        # - get actual positions from exchange
        for instr in instruments:
            self.__cache.init_ohlcv(instr)

        # - subscribe to market data
        self.__subscription_manager.subscribe(instruments)

        # - reinitialize strategy loggers
        self.__logging.initialize(
            self.__time_provider.time(), self.__positions, self.__trading_service.get_account().get_balances()
        )

    def _create_and_update_positions(self, instruments: list[Instrument]):
        for instrument in instruments:
            _ = self.__trading_service.get_position(instrument)

            # - check if we need any aux instrument for calculating pnl ?
            # TODO: test edge cases for aux symbols
            aux = lookup.find_aux_instrument_for(instrument, self.__trading_service.get_base_currency())
            if aux is not None:
                instrument._aux_instrument = aux
                instruments.append(aux)
                _ = self.__trading_service.get_position(aux)



import pandas as pd
import numpy as np

from typing import Any, Dict, List, Optional, Tuple

from qubx import logger
from qubx.core.basics import Order, Deal, Position, Instrument
from qubx.core.series import TimeSeries, Bar, Trade, Quote, OrderBook, time_as_nsec
from qubx.utils.orderbook import build_orderbook_snapshots
from .ccxt_exceptions import CcxtOrderBookParsingError


EXCHANGE_ALIASES = {"binance.um": "binanceusdm", "binance.cm": "binancecoinm", "kraken.f": "krakenfutures"}

DATA_PROVIDERS_ALIASES = EXCHANGE_ALIASES | {"binance": "binanceqv", "binance.um": "binanceqv_usdm"}


def ccxt_convert_order_info(instrument: Instrument, raw: dict[str, Any]) -> Order:
    """
    Convert CCXT excution record to Order object
    """
    ri = raw["info"]
    amnt = float(ri.get("origQty", raw.get("amount")))
    price = raw["price"]
    status = raw["status"]
    side = raw["side"].upper()
    _type = ri.get("type", raw.get("type")).upper()
    if status == "open":
        status = ri.get("status", status)  # for filled / part_filled ?

    return Order(
        id=raw["id"],
        type=_type,
        instrument=instrument,
        time=pd.Timestamp(raw["timestamp"], unit="ms"),  # type: ignore
        quantity=amnt,
        price=float(price) if price is not None else 0.0,
        side=side,
        status=status.upper(),
        time_in_force=raw["timeInForce"],
        client_id=raw["clientOrderId"],
        cost=float(raw["cost"] or 0),  # cost can be None
    )


def ccxt_convert_deal_info(raw: Dict[str, Any]) -> Deal:
    fee_amount = None
    fee_currency = None
    if "fee" in raw:
        fee_amount = float(raw["fee"]["cost"])
        fee_currency = raw["fee"]["currency"]
    return Deal(
        id=raw["id"],
        order_id=raw["order"],
        time=pd.Timestamp(raw["timestamp"], unit="ms"),  # type: ignore
        amount=float(raw["amount"]) * (-1 if raw["side"] == "sell" else +1),
        price=float(raw["price"]),
        aggressive=raw["takerOrMaker"] == "taker",
        fee_amount=fee_amount,
        fee_currency=fee_currency,
    )


def ccxt_extract_deals_from_exec(report: Dict[str, Any]) -> List[Deal]:
    """
    Small helper for extracting deals (trades) from CCXT execution report
    """
    deals = list()
    if trades := report.get("trades"):
        for t in trades:
            deals.append(ccxt_convert_deal_info(t))
    return deals


def ccxt_restore_position_from_deals(
    pos: Position, current_volume: float, deals: List[Deal], reserved_amount: float = 0.0
) -> Position:
    if current_volume != 0:
        instr = pos.instrument
        _last_deals = []

        # - try to find last deals that led to this position
        for d in sorted(deals, key=lambda x: x.time, reverse=True):
            current_volume -= d.amount
            # - spot case when fees may be deducted from the base coin
            #   that may decrease total amount
            if d.fee_amount is not None:
                if instr.base == d.fee_currency:
                    current_volume += d.fee_amount
            # print(d.amount, current_volume)
            _last_deals.insert(0, d)

            # - take in account reserves
            if abs(current_volume) - abs(reserved_amount) < instr.min_size_step:
                break

        # - reset to 0
        pos.reset()

        if abs(current_volume) - abs(reserved_amount) > instr.min_size_step:
            # - - - TODO - - - !!!!
            logger.warning(
                f"Couldn't restore full deals history for {instr.symbol} symbol. Qubx will use zero position !"
            )
        else:
            fees_in_base = 0.0
            for d in _last_deals:
                pos.update_position_by_deal(d)
                if d.fee_amount is not None:
                    if instr.base == d.fee_currency:
                        fees_in_base += d.fee_amount
            # - we round fees up in case of fees in base currency
            pos.quantity -= pos.instrument.round_size_up(fees_in_base)
    return pos


def ccxt_convert_trade(trade: dict[str, Any]) -> Trade:
    t_ns = trade["timestamp"] * 1_000_000  # this is trade time
    s, info, price, amnt = trade["symbol"], trade["info"], trade["price"], trade["amount"]
    m = info["m"]
    return Trade(t_ns, price, amnt, int(not m), int(trade["id"]))


def ccxt_convert_orderbook(
    ob: dict, instr: Instrument, levels: int = 50, tick_size_pct: float = 0.01, sizes_in_quoted: bool = False
) -> OrderBook | None:
    """
    Convert a ccxt order book to an OrderBook object with a fixed tick size percentage.
    Parameters:
        ob (dict): The order book dictionary from ccxt.
        instr (Instrument): The instrument object containing market-specific details.
        levels (int, optional): The number of levels to include in the order book. Default is 50.
        tick_size_pct (float, optional): The tick size percentage. Default is 0.01%.
        sizes_in_quoted (bool, optional): Whether the size is in the quoted currency. Default is False.
    Returns:
        OrderBook: The converted OrderBook object.
    """
    _dt = pd.Timestamp(ob["datetime"]).replace(tzinfo=None).asm8
    _prev_dt = _dt - pd.Timedelta("1ms").asm8

    updates = [
        *[(_prev_dt, update[0], update[1], True) for update in ob["bids"]],
        *[(_prev_dt, update[0], update[1], False) for update in ob["asks"]],
    ]
    # add an artificial update to trigger the snapshot building
    updates.append((_dt, 0, 0, True))

    try:
        snapshots = build_orderbook_snapshots(
            updates,
            levels=levels,
            tick_size_pct=tick_size_pct,
            min_tick_size=instr.min_tick,
            min_size_step=instr.min_size_step,
            sizes_in_quoted=sizes_in_quoted,
        )
    except Exception as e:
        logger.error(f"Failed to build order book snapshots: {e}")
        snapshots = None

    if not snapshots:
        return None

    (dt, _bids, _asks, top_bid, top_ask, tick_size) = snapshots[-1]
    bids = np.array([s for _, s in _bids[::-1]])
    asks = np.array([s for _, s in _asks])

    return OrderBook(
        time=time_as_nsec(dt),
        top_bid=top_bid,
        top_ask=top_ask,
        tick_size=tick_size,
        bids=bids,
        asks=asks,
    )



from typing import Any, Optional, List, Dict, Tuple
from collections import defaultdict

import numpy as np

from qubx import logger
from qubx.core.basics import Instrument, Position, dt_64, Deal, Order


class AccountProcessor:
    """
    Account processor class
    """

    account_id: str
    base_currency: str
    reserved: Dict[str, float]  # how much asset is reserved against the trading
    _balances: Dict[str, Tuple[float, float]]
    _active_orders: Dict[str | int, Order]  # active orders
    _processed_trades: Dict[str | int, List[str | int]]
    _positions: Dict[Instrument, Position]
    _locked_capital_by_order: Dict[str | int, float]

    def __init__(
        self,
        account_id: str,
        base_currency: str,
        reserves: Dict[str, float] | None = None,
        initial_capital: float = 100_000,
    ) -> None:
        self.account_id = account_id
        self.base_currency = base_currency.upper()
        self.reserved = dict() if reserves is None else reserves
        self._processed_trades = defaultdict(list)
        self._active_orders = dict()
        self._positions = {}
        self._locked_capital_by_order = dict()
        self._balances = dict()
        self._balances[self.base_currency] = (initial_capital, 0)

    @property
    def positions(self) -> dict[Instrument, Position]:
        return self._positions

    def update_balance(self, currency: str, total_capital: float, locked_capital: float):
        if currency == self.base_currency:
            current_balance = self._balances[currency][0]
            if current_balance > 0 and current_balance < total_capital:
                logger.info(
                    f"Restricting {self.base_currency} capital from {total_capital} "
                    f"down to {current_balance} due to initial capital constraint."
                )
                total_capital = current_balance
        self._balances[currency] = (total_capital, locked_capital)

    def get_balances(self) -> Dict[str, Tuple[float, float]]:
        return dict(self._balances)

    def attach_positions(self, *position: Position) -> "AccountProcessor":
        for p in position:
            self._positions[p.instrument] = p
        return self

    def update_position_price(self, time: dt_64, instrument: Instrument, price: float):
        p = self._positions[instrument]
        p.update_market_price(time, price, 1)

    def get_capital(self) -> float:
        """
        Get free capital in base currency.
        """
        # TODO: separate cash and margin accounts in the future
        return self.base_currency_balance - self.base_currency_locked_balance

    def get_total_capital(self) -> float:
        """
        Get margin capital in base currency. It includes capital locked in positions.
        """
        positions_value_in_base = sum([p.market_value_funds for p in self._positions.values()])
        return self.base_currency_balance + positions_value_in_base - self.get_reserved_capital()

    def get_leverage(self, instrument: Instrument) -> float:
        pos = self._positions.get(instrument)
        if pos is not None:
            return pos.market_value_funds / self.get_total_capital()
        return 0.0

    def get_leverages(self) -> dict[Instrument, float]:
        return {s: self.get_leverage(s) for s in self._positions.keys()}

    def get_net_leverage(self) -> float:
        return sum(self.get_leverages().values())

    def get_gross_leverage(self) -> float:
        return sum(map(abs, self.get_leverages().values()))

    def get_reserved(self, instrument: Instrument) -> float:
        """
        Check how much were reserved for this instrument
        """
        return self.reserved.get(instrument.base, 0.0)

    def get_reserved_capital(self) -> float:
        """
        Get total amount of capital reserved from trading.
        """
        if not self.reserved:
            return 0.0
        reserved_capital = 0.0
        for currency, amount in self.reserved.items():
            instrument = self._find_instrument_for_currency(currency)
            reserved_capital += amount * self._positions[instrument].last_update_price
        return reserved_capital

    def process_deals(self, instrument: Instrument, deals: list[Deal]):
        pos = self._positions.get(instrument)

        if pos is not None:
            conversion_rate = 1
            instr = pos.instrument
            traded_amnt, realized_pnl, deal_cost = 0, 0, 0

            # - check if we need conversion rate for this instrument
            # - TODO - need test on it !
            if instr._aux_instrument is not None:
                aux_pos = self._positions.get(instr._aux_instrument)
                if aux_pos:
                    conversion_rate = aux_pos.last_update_price
                else:
                    logger.error(
                        f"Can't find additional instrument {instr._aux_instrument} for estimating {instrument.symbol} position value !!!"
                    )

            # - process deals
            for d in deals:
                if d.id not in self._processed_trades[d.order_id]:
                    self._processed_trades[d.order_id].append(d.id)
                    r_pnl, fee_in_base = pos.update_position_by_deal(d, conversion_rate)
                    realized_pnl += r_pnl
                    deal_cost += d.amount * d.price / conversion_rate
                    traded_amnt += d.amount
                    logger.debug(f"  ::  traded {d.amount} for {instrument} @ {d.price} -> {realized_pnl:.2f}")
                    self.base_currency_balance -= deal_cost + fee_in_base

    def _lock_limit_order_value(self, order: Order) -> float:
        pos = self._positions.get(order.instrument)
        excess = 0.0
        # - we handle only instruments it;s subscribed to
        if pos:
            sgn = -1 if order.side == "SELL" else +1
            pos_change = sgn * order.quantity
            direction = np.sign(pos_change)
            prev_direction = np.sign(pos.quantity)
            # how many shares are closed/open
            qty_closing = min(abs(pos.quantity), abs(pos_change)) * direction if prev_direction != direction else 0
            qty_opening = pos_change if prev_direction == direction else pos_change - qty_closing
            excess = abs(qty_opening) * order.price

            if excess > 0:
                self.base_currency_locked_balance += excess
                self._locked_capital_by_order[order.id] = excess

        return excess

    def _unlock_limit_order_value(self, order: Order):
        if order.id in self._locked_capital_by_order:
            excess = self._locked_capital_by_order.pop(order.id)
            self.base_currency_locked_balance -= excess

    def process_order(self, order: Order):
        _new = order.status == "NEW"
        _open = order.status == "OPEN"
        _closed = order.status == "CLOSED"
        _cancel = order.status == "CANCELED"

        if _open or _new:
            self._active_orders[order.id] = order

            # - calculate amount locked by this order
            if order.type == "LIMIT":
                self._lock_limit_order_value(order)

        if _closed or _cancel:
            if order.id in self._processed_trades:
                self._processed_trades.pop(order.id)

            if order.id in self._active_orders:
                self._active_orders.pop(order.id)

        # - calculate amount to unlock after canceling
        if _cancel and order.type == "LIMIT":
            self._unlock_limit_order_value(order)

        logger.debug(
            f"Order {order.id} {order.type} {order.side} {order.quantity} of {order.instrument} -> {order.status}"
        )

    def add_active_orders(self, orders: Dict[str, Order]):
        for oid, od in orders.items():
            self._active_orders[oid] = od

    def get_orders(self, instrument: Instrument | None = None) -> List[Order]:
        ols = list(self._active_orders.values())
        if instrument is not None:
            ols = list(filter(lambda x: x.instrument == instrument, ols))
        return ols

    def positions_report(self) -> dict:
        rep = {}
        for p in self._positions.values():
            rep[p.instrument.symbol] = {
                "Qty": p.quantity,
                "Price": p.position_avg_price_funds,
                "PnL": p.pnl,
                "MktValue": p.market_value_funds,
                "Leverage": self.get_leverage(p.instrument),
            }
        return rep

    @property
    def base_currency_balance(self) -> float:
        return self._balances[self.base_currency][0]

    @property
    def base_currency_locked_balance(self) -> float:
        return self._balances[self.base_currency][1]

    @base_currency_balance.setter
    def base_currency_balance(self, value: float):
        self._balances[self.base_currency] = (value, self._balances[self.base_currency][1])

    @base_currency_locked_balance.setter
    def base_currency_locked_balance(self, value: float):
        self._balances[self.base_currency] = (self._balances[self.base_currency][0], value)

    def _find_instrument_for_currency(self, currency: str) -> Instrument:
        for instr in self._positions.keys():
            if instr.base == currency and instr.quote == self.base_currency:
                return instr
        raise ValueError(f"Can't find instrument for currency {currency}")



import asyncio
import re
import numpy as np
import pandas as pd
import ccxt.pro as cxp
import concurrent.futures

from threading import Event, Lock, Thread
from typing import Any, Dict, List, Optional, Callable, Awaitable, Tuple, Set
from asyncio.tasks import Task
from asyncio.events import AbstractEventLoop
from asyncio.exceptions import CancelledError
from collections import defaultdict
from types import FunctionType
from ccxt import NetworkError, ExchangeClosedByUser, ExchangeError, ExchangeNotAvailable
from ccxt.base.exchange import Exchange

from qubx import logger
from qubx.core.basics import Instrument, Position, dt_64, Deal, CtrlChannel, SubscriptionType
from qubx.core.helpers import BasicScheduler
from qubx.core.interfaces import IBrokerServiceProvider, ITradingServiceProvider, ITimeProvider
from qubx.utils.threading import synchronized
from qubx.core.series import TimeSeries, Bar, Trade, Quote
from qubx.utils.ntp import start_ntp_thread, time_now
from .ccxt_utils import DATA_PROVIDERS_ALIASES, ccxt_convert_trade, ccxt_convert_orderbook
from .ccxt_exceptions import CcxtSymbolNotRecognized

# - register custom wrappers
from .ccxt_customizations import BinanceQV, BinanceQVUSDM

cxp.binanceqv = BinanceQV  # type: ignore
cxp.binanceqv_usdm = BinanceQVUSDM  # type: ignore
cxp.exchanges.append("binanceqv")
cxp.exchanges.append("binanceqv_usdm")


EXCH_SYMBOL_PATTERN = re.compile(r"(?P<base>[^/]+)/(?P<quote>[^:]+)(?::(?P<margin>.+))?")


class ProxyCtrlChannel(CtrlChannel):
    """
    ProxyCtrlChannel is a wrapper around CtrlChannel that tracks the time of the last message sent.
    """

    control: Event
    name: str
    lock: Lock

    _channel: CtrlChannel
    _time_provider: ITimeProvider
    _sub_instr_to_time: Dict[Tuple[str, Instrument], dt_64]

    def __init__(
        self, channel: CtrlChannel, time_provider: ITimeProvider, sub_instr_to_time: Dict[Tuple[str, Instrument], dt_64]
    ):
        self.name = channel.name
        self.control = channel.control
        self.lock = channel.lock
        self._channel = channel
        self._time_provider = time_provider
        self._sub_instr_to_time = sub_instr_to_time

    def send(self, msg: Tuple):
        self._channel.send(msg)
        if len(msg) != 3:
            raise ValueError(f"Invalid message {msg}")
        instr, sub_type, _ = msg
        self._sub_instr_to_time[(sub_type, instr)] = self._time_provider.time()


class CCXTExchangesConnector(IBrokerServiceProvider):

    _exchange: Exchange
    _scheduler: BasicScheduler | None = None

    # - subscriptions
    _subscriptions: Dict[str, Set[Instrument]]
    _subscription_to_params: Dict[str, Dict[str, Any]]
    _pending_stream_subscriptions: Dict[Tuple[str, Tuple], Set[Instrument]]
    _pending_stream_unsubscriptions: Dict[str, Set[Instrument]]
    _pending_instrument_unsubscriptions: Set[Instrument]
    _sub_to_coro: Dict[str, concurrent.futures.Future]
    _sub_to_name: Dict[str, str]
    _is_sub_name_enabled: Dict[str, bool]

    _sub_instr_to_time: Dict[Tuple[str, Instrument], dt_64]
    _last_quotes: Dict[str, Optional[Quote]]
    _loop: AbstractEventLoop
    _thread_event_loop: Thread
    _subscribers: Dict[str, Callable]

    def __init__(
        self,
        exchange_id: str,
        trading_service: ITradingServiceProvider,
        read_only: bool = False,
        loop: AbstractEventLoop | None = None,
        max_ws_retries: int = 10,
        **exchange_auth,
    ):
        super().__init__(exchange_id, trading_service)
        self.trading_service = trading_service
        self.read_only = read_only
        self.max_ws_retries = max_ws_retries
        exchange_id = exchange_id.lower()

        # - start NTP thread
        start_ntp_thread()

        # - setup communication bus
        self.set_communication_channel(bus := CtrlChannel("databus", sentinel=(None, None, None)))
        self.trading_service.set_communication_channel(bus)

        # - init CCXT stuff
        exch = DATA_PROVIDERS_ALIASES.get(exchange_id, exchange_id)
        if exch not in cxp.exchanges:
            raise ValueError(f"Exchange {exchange_id} -> {exch} is not supported by CCXT.pro !")

        # - create new even loop
        self._loop = asyncio.new_event_loop() if loop is None else loop
        asyncio.set_event_loop(self._loop)

        # - create exchange's instance
        self._exchange = getattr(cxp, exch)(exchange_auth | {"asyncio_loop": self._loop})
        self._last_quotes = defaultdict(lambda: None)
        self._subscriptions = defaultdict(set)
        self._subscription_to_params = defaultdict(dict)
        self._pending_stream_subscriptions = defaultdict(set)
        self._pending_stream_unsubscriptions = defaultdict(set)
        self._pending_instrument_unsubscriptions = set()
        self._sub_to_coro = {}
        self._sub_to_name = {}
        self._is_sub_name_enabled = defaultdict(lambda: False)
        self._sub_instr_to_time = {}

        self._subscribers = {
            n.split("_subscribe_")[1]: f
            for n, f in self.__class__.__dict__.items()
            if type(f) == FunctionType and n.startswith("_subscribe_")
        }

        logger.info(f"{exchange_id} initialized - current time {self.trading_service.time()}")

    @property
    def is_simulated_trading(self) -> bool:
        return False

    def get_scheduler(self) -> BasicScheduler:
        if self._scheduler is None:
            self._scheduler = BasicScheduler(self.get_communication_channel(), lambda: self.time().item())
        return self._scheduler

    def time(self) -> dt_64:
        return time_now()

    def subscribe(
        self,
        instruments: List[Instrument],
        subscription_type: str,
        **kwargs,
    ) -> None:
        self._pending_stream_subscriptions[
            self._parse_subscription_type(subscription_type), tuple(kwargs.items())
        ].update(instruments)

    def unsubscribe(self, instruments: List[Instrument], subscription_type: str | None = None) -> None:
        if subscription_type is None:
            self._pending_instrument_unsubscriptions.update(instruments)
        else:
            self._pending_stream_unsubscriptions[self._parse_subscription_type(subscription_type)].update(instruments)

    def get_subscriptions(self, instrument: Instrument) -> Dict[str, Dict[str, Any]]:
        _sub_to_params = {}
        for sub, instrs in self._subscriptions.items():
            if instrument in instrs:
                _sub_to_params[sub] = self._subscription_to_params[sub]
        return _sub_to_params

    def has_subscription(self, subscription_type: str, instrument: Instrument) -> bool:
        sub = subscription_type.lower()
        return sub in self._subscriptions and instrument in self._subscriptions[sub]

    @synchronized
    def commit(self) -> None:
        _current_instruments = self.subscribed_instruments
        _sub_to_params = self._get_updated_sub_to_params()
        _has_something_changed = False
        for sub, params in _sub_to_params.items():
            _current_instruments = self._subscriptions[sub]
            _removed_instruments = self._get_pending_unsub_instr(sub)
            _added_instruments = self._get_pending_sub_instr(sub)
            _updated_instruments = _current_instruments.union(_added_instruments).difference(_removed_instruments)
            if _updated_instruments != _current_instruments:
                _has_something_changed = True
                self._subscribe(_updated_instruments, sub, **params)

        if _has_something_changed and not self.read_only:
            _new_instruments = self.subscribed_instruments
            _removed_instruments = _current_instruments.difference(_new_instruments)
            _added_instruments = _new_instruments.difference(_current_instruments)
            self._update_execution_subscriptions(_added_instruments, _removed_instruments)

        # - clean up pending subs and unsubs
        self._pending_stream_subscriptions.clear()
        self._pending_instrument_unsubscriptions.clear()
        self._pending_stream_unsubscriptions.clear()

    def get_quote(self, instrument: Instrument) -> Quote | None:
        return self._last_quotes[instrument.symbol]

    def get_historical_ohlcs(self, instrument: Instrument, timeframe: str, nbarsback: int) -> List[Bar]:
        assert nbarsback >= 1
        symbol = instrument.symbol
        since = self._time_msec_nbars_back(timeframe, nbarsback)

        # - retrieve OHLC data
        # - TODO: check if nbarsback > max_limit (1000) we need to do more requests
        # - TODO: how to get quoted volumes ?
        async def _get():
            return await self._exchange.fetch_ohlcv(
                symbol, self._get_exch_timeframe(timeframe), since=since, limit=nbarsback + 1
            )  # type: ignore

        fut = asyncio.run_coroutine_threadsafe(_get(), self._loop)
        res = fut.result(60)

        _arr = []
        for oh in res:  # type: ignore
            _arr.append(
                Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7])
                if len(oh) > 6
                else Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[5])
            )

        return _arr

    def close(self):
        try:
            if hasattr(self._exchange, "close"):
                future = self._submit_coro(self._exchange.close())  # type: ignore
                # - wait for 5 seconds for connection to close
                future.result(5)
            else:
                del self._exchange
        except Exception as e:
            logger.error(e)

    @property
    def subscribed_instruments(self) -> Set[Instrument]:
        if not self._subscriptions:
            return set()
        return set.union(*self._subscriptions.values())

    def _get_pending_unsub_instr(self, sub: str) -> Set[Instrument]:
        return self._pending_stream_unsubscriptions.get(sub, set()).union(self._pending_instrument_unsubscriptions)

    def _get_pending_sub_instr(self, sub: str) -> Set[Instrument]:
        for (sub_type, _), instrs in self._pending_stream_subscriptions.items():
            if sub_type == sub:
                return instrs
        return set()

    def _subscribe(
        self,
        instruments: Set[Instrument],
        sub_type: str,
        **kwargs,
    ) -> None:
        _subscriber = self._subscribers.get(sub_type)
        if _subscriber is None:
            raise ValueError(f"Subscription type {sub_type} is not supported")

        self._check_event_loop_is_running()

        _current_instruments = self._subscriptions[sub_type]
        _added_instruments = instruments.difference(_current_instruments)

        _channel = self._get_proxy_channel()

        # - fetch historical ohlc for added instruments
        if (ohlc_warmup_period := kwargs.get("ohlc_warmup_period")) is not None and (
            timeframe := kwargs.get("timeframe")
        ) is not None:
            self._run_ohlc_warmup_sync(_channel, _added_instruments, timeframe, ohlc_warmup_period, timeout=60 * 10)

        self._resubscribe_stream(sub_type, _channel, instruments, **kwargs)
        self._subscriptions[sub_type] = instruments
        self._subscription_to_params[sub_type] = kwargs

    def _stop_subscriber(self, sub_type: str) -> None:
        if sub_type not in self._sub_to_coro:
            return
        sub_name = self._sub_to_name[sub_type]
        self._is_sub_name_enabled[sub_name] = False  # stop the subscriber
        future = self._sub_to_coro[sub_type]
        future.result(10)  # wait for 10 seconds for the future to finish
        if future.running():
            future.cancel()
        del self._sub_to_coro[sub_type]
        del self._sub_to_name[sub_type]
        del self._is_sub_name_enabled[sub_name]

    def _get_proxy_channel(self) -> ProxyCtrlChannel:
        return ProxyCtrlChannel(self.get_communication_channel(), self, self._sub_instr_to_time)

    def _resubscribe_stream(self, sub_type: str, channel: CtrlChannel, instruments: Set[Instrument], **kwargs) -> None:
        if sub_type in self._sub_to_coro:
            logger.debug(f"Canceling existing {sub_type} subscription for {self._subscriptions[sub_type]}")
            self._stop_subscriber(sub_type)

        if not instruments:
            return

        _subscriber = self._subscribers[sub_type]
        _subscriber_params = set(_subscriber.__code__.co_varnames[: _subscriber.__code__.co_argcount])
        # - get only parameters that are needed for subscriber
        kwargs = {k: v for k, v in kwargs.items() if k in _subscriber_params}
        self._sub_to_name[sub_type] = (name := self._get_subscription_name(instruments, sub_type, **kwargs))
        self._sub_to_coro[sub_type] = self._submit_coro(
            _subscriber(self, name, sub_type, channel, instruments, **kwargs)
        )

    def _check_event_loop_is_running(self) -> None:
        if self._loop.is_running():
            return
        self._thread_event_loop = Thread(target=self._loop.run_forever, args=(), daemon=True)
        self._thread_event_loop.start()

    def _parse_subscription_type(self, subscription_type: str) -> str:
        return subscription_type.lower()

    def _get_updated_sub_to_params(self) -> Dict[str, Dict[str, Any]]:
        _update_subs = set(self._pending_stream_unsubscriptions.keys())
        if self._pending_instrument_unsubscriptions:
            _update_subs |= set(self._subscriptions.keys())
        _update_subs.update(sub for sub, _ in self._pending_stream_subscriptions)
        _sub_to_params = {
            **self._subscription_to_params,
            **{sub: dict(sub_args) for sub, sub_args in self._pending_stream_subscriptions.keys()},
        }
        _sub_to_params = {sub: _sub_to_params.get(sub, {}) for sub in _update_subs}
        return _sub_to_params

    def _submit_coro(self, coro: Awaitable[None]) -> concurrent.futures.Future:
        return asyncio.run_coroutine_threadsafe(coro, self._loop)

    def _time_msec_nbars_back(self, timeframe: str, nbarsback: int = 1) -> int:
        return (self.time() - nbarsback * pd.Timedelta(timeframe)).asm8.item() // 1000000

    def _run_ohlc_warmup_sync(
        self, channel: CtrlChannel, instruments: Set[Instrument], timeframe: str, warmup_period: str, timeout: int
    ) -> None:
        future = self._submit_coro(self._run_ohlc_warmup(channel, instruments, timeframe, warmup_period))
        future.result(timeout)

    async def _run_ohlc_warmup(
        self, channel: CtrlChannel, instruments: Set[Instrument], timeframe: str, warmup_period: str
    ) -> None:
        if not instruments:
            return
        logger.debug(f"Running OHLC warmup for {instruments} with period {warmup_period}")
        await asyncio.gather(
            *[self._fetch_and_send_ohlc(channel, instr, timeframe, warmup_period) for instr in instruments]
        )

    async def _listen_to_stream(
        self,
        subscriber: Callable[[], Awaitable[None]],
        exchange: Exchange,
        channel: CtrlChannel,
        name: str,
        unsubscriber: Callable[[], Awaitable[None]] | None = None,
    ):
        logger.debug(f"Listening to {name}")
        self._is_sub_name_enabled[name] = True
        n_retry = 0
        while channel.control.is_set():
            try:
                await subscriber()
                n_retry = 0
                if not self._is_sub_name_enabled[name]:
                    if unsubscriber is not None:
                        await unsubscriber()
                    break
            except CancelledError:
                if unsubscriber is not None:
                    try:
                        # - unsubscribe from stream, but ignore if there is an error
                        # because we anyway close the connection
                        await unsubscriber()
                    except:
                        pass
                break
            except ExchangeClosedByUser:
                # - we closed connection so just stop it
                logger.info(f"{name} listening has been stopped")
                break
            except (NetworkError, ExchangeError, ExchangeNotAvailable) as e:
                logger.error(f"Error in {name} : {e}")
                await asyncio.sleep(1)
                continue
            except Exception as e:
                logger.error(f"exception in {name} : {e}")
                logger.exception(e)
                n_retry += 1
                if n_retry >= self.max_ws_retries:
                    logger.error(f"Max retries reached for {name}. Closing connection.")
                    del exchange
                    break
                await asyncio.sleep(min(2**n_retry, 60))  # Exponential backoff with a cap at 60 seconds

    async def _fetch_and_send_ohlc(
        self, channel: CtrlChannel, instrument: Instrument, timeframe: str, warmup_period: str
    ) -> None:
        nbarsback = pd.Timedelta(warmup_period) // pd.Timedelta(timeframe)
        exch_timeframe = self._get_exch_timeframe(timeframe)
        start = self._time_msec_nbars_back(timeframe, nbarsback)
        ohlcv = await self._exchange.fetch_ohlcv(instrument.symbol, exch_timeframe, since=start, limit=nbarsback + 1)
        # - just send data as the list
        channel.send(
            (
                instrument,
                self._get_hist_type(SubscriptionType.OHLC),
                [Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7]) for oh in ohlcv],
            )
        )
        logger.info(f"{instrument}: loaded {len(ohlcv)} {timeframe} bars")

    def _get_exch_timeframe(self, timeframe: str) -> str:
        if timeframe is not None:
            _t = re.match(r"(\d+)(\w+)", timeframe)
            timeframe = f"{_t[1]}{_t[2][0].lower()}" if _t and len(_t.groups()) > 1 else timeframe

        tframe = self._exchange.find_timeframe(timeframe)
        if tframe is None:
            raise ValueError(f"timeframe {timeframe} is not supported by {self._exchange.name}")

        return tframe

    def _get_exch_symbol(self, instrument: Instrument) -> str:
        return f"{instrument.base}/{instrument.quote}:{instrument.margin_symbol}"

    def _find_instrument_for_exch_symbol(
        self, exch_symbol: str, symbol_to_instrument: Dict[str, Instrument]
    ) -> Instrument:
        match = EXCH_SYMBOL_PATTERN.match(exch_symbol)
        if not match:
            raise CcxtSymbolNotRecognized(f"Invalid exchange symbol {exch_symbol}")
        base = match.group("base")
        quote = match.group("quote")
        symbol = f"{base}{quote}"
        if symbol not in symbol_to_instrument:
            raise CcxtSymbolNotRecognized(f"Unknown symbol {symbol}")
        return symbol_to_instrument[symbol]

    def _get_subscription_name(
        self, instruments: List[Instrument] | Set[Instrument] | Instrument, subscription: str, **kwargs
    ) -> str:
        if isinstance(instruments, Instrument):
            instruments = [instruments]
        _symbols = [i.symbol for i in instruments]
        _name = f"{','.join(_symbols)} {subscription}"
        if kwargs:
            kwargs_str = ",".join(f"{k}={v}" for k, v in kwargs.items())
            _name += f" ({kwargs_str})"
        return _name

    def _get_hist_type(self, sub_type: str) -> str:
        return f"hist_{sub_type}"

    def _update_execution_subscriptions(self, added_instruments: Set[Instrument], removed_instruments: Set[Instrument]):
        channel = self._get_proxy_channel()
        _get_name = lambda instr: self._get_subscription_name(instr, "executions")
        # - subscribe to added instruments
        for instr in added_instruments:
            _name = _get_name(instr)
            future = self._sub_to_coro.get(_name)
            if future is not None and future.running():
                logger.warning(f"Execution subscription for {instr} is already running")
                continue
            self._sub_to_coro[_name] = self._submit_coro(
                self._subscribe_executions(_name, "executions", channel, instr)
            )
        # - unsubscribe from removed instruments
        for instr in removed_instruments:
            _name = _get_name(instr)
            self._is_sub_name_enabled[_name] = False
            future = self._sub_to_coro[_name]
            try:
                # usually this will raise a timeout error because there will be no order updates
                # so just cancel
                future.result(0.1)
            except TimeoutError:
                if future.running():
                    future.cancel()

    def _get_latest_instruments(self, sub_type: str, warmup_period: str) -> Set[Instrument]:
        """
        Get the latest instruments that have relatively fresh data that does not exceed 1/4 of the warmup period.
        """
        _oldest_allowed_time = (pd.Timestamp(self.time()) - pd.Timedelta(warmup_period) / 4).asm8
        return {
            instr
            for (_sub_type, instr), _time in self._sub_instr_to_time.items()
            if _sub_type == sub_type and _time > _oldest_allowed_time
        }

    #############################
    # - Subscription methods
    #############################
    async def _subscribe_ohlc(
        self,
        name: str,
        sub_type: str,
        channel: CtrlChannel,
        instruments: Set[Instrument],
        warmup_period: str | None = None,
        timeframe: str = "1m",
    ):
        symbols = [i.symbol for i in instruments]
        _exchange_timeframe = self._get_exch_timeframe(timeframe)
        _symbol_timeframe_pairs = [[symbol, _exchange_timeframe] for symbol in symbols]
        _symbol_to_instrument = {i.symbol: i for i in instruments}

        async def watch_ohlcv():
            ohlcv = await self._exchange.watch_ohlcv_for_symbols(_symbol_timeframe_pairs)
            _time = self.time()
            # - ohlcv is symbol -> timeframe -> list[timestamp, open, high, low, close, volume]
            for exch_symbol, _data in ohlcv.items():
                instrument = self._find_instrument_for_exch_symbol(exch_symbol, _symbol_to_instrument)
                for _, ohlcvs in _data.items():
                    for oh in ohlcvs:
                        channel.send(
                            (
                                instrument,
                                sub_type,
                                Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7]),
                            )
                        )
                    last_close = ohlcvs[-1][4]
                    self.trading_service.update_position_price(
                        instrument,
                        _time,
                        last_close,
                    )

        async def un_watch_ohlcv():
            unwatch = getattr(self._exchange, "un_watch_ohlcv_for_symbols", lambda _: None)(_symbol_timeframe_pairs)
            if unwatch is not None:
                await unwatch

        if warmup_period is not None:
            _new_instruments = instruments.difference(self._get_latest_instruments(sub_type, warmup_period))
            await self._run_ohlc_warmup(channel, _new_instruments, timeframe, warmup_period)

        await self._listen_to_stream(
            subscriber=watch_ohlcv,
            exchange=self._exchange,
            channel=channel,
            name=name,
            unsubscriber=un_watch_ohlcv,
        )

    async def _subscribe_trade(
        self,
        name: str,
        sub_type: str,
        channel: CtrlChannel,
        instruments: Set[Instrument],
        warmup_period: str | None = None,
    ):
        symbols = [i.symbol for i in instruments]
        _symbol_to_instrument = {i.symbol: i for i in instruments}

        async def watch_trades():
            trades = await self._exchange.watch_trades_for_symbols(symbols)
            symbol = trades[0]["symbol"]
            _time = self.time()
            instrument = self._find_instrument_for_exch_symbol(symbol, _symbol_to_instrument)
            last_trade = ccxt_convert_trade(trades[-1])
            self.trading_service.update_position_price(instrument, _time, last_trade)
            for trade in trades:
                channel.send((instrument, sub_type, ccxt_convert_trade(trade)))

        async def un_watch_trades():
            unwatch = getattr(self._exchange, "un_watch_trades_for_symbols", lambda _: None)(symbols)
            if unwatch is not None:
                await unwatch

        if warmup_period is not None:
            _new_instruments = instruments.difference(self._get_latest_instruments(sub_type, warmup_period))
            if _new_instruments:
                logger.debug(f"New instruments for trade warmup: {_new_instruments}")

                async def fetch_and_send_trades(instr: Instrument):
                    trades = await self._exchange.fetch_trades(
                        instr.symbol, since=self._time_msec_nbars_back(warmup_period)
                    )
                    logger.debug(f"Loaded {len(trades)} trades for {instr}")
                    channel.send(
                        (
                            instr,
                            self._get_hist_type(sub_type),
                            [ccxt_convert_trade(trade) for trade in trades],
                        )
                    )

                await asyncio.gather(*(fetch_and_send_trades(instr) for instr in _new_instruments))

        await self._listen_to_stream(
            subscriber=watch_trades,
            exchange=self._exchange,
            channel=channel,
            name=name,
            unsubscriber=un_watch_trades,
        )

    async def _subscribe_orderbook(
        self,
        name: str,
        sub_type: str,
        channel: CtrlChannel,
        instruments: Set[Instrument],
        warmup_period: str | None = None,
    ):
        symbols = [i.symbol for i in instruments]
        _symbol_to_instrument = {i.symbol: i for i in instruments}

        async def watch_orderbook():
            ccxt_ob = await self._exchange.watch_order_book_for_symbols(symbols)
            exch_symbol = ccxt_ob["symbol"]
            instrument = self._find_instrument_for_exch_symbol(exch_symbol, _symbol_to_instrument)
            ob = ccxt_convert_orderbook(ccxt_ob, instrument)
            if ob is None:
                return
            quote = ob.to_quote()
            self._last_quotes[instrument.symbol] = quote
            self.trading_service.update_position_price(instrument, self.time(), quote)
            channel.send((instrument, sub_type, ob))

        async def un_watch_orderbook():
            unwatch = getattr(self._exchange, "un_watch_order_book_for_symbols", lambda _: None)(symbols)
            if unwatch is not None:
                await unwatch

        # TODO: add historical orderbook fetching

        await self._listen_to_stream(
            subscriber=watch_orderbook,
            exchange=self._exchange,
            channel=channel,
            name=name,
            unsubscriber=un_watch_orderbook,
        )

    async def _subscribe_executions(self, name: str, sub_type: str, channel: CtrlChannel, instrument: Instrument):
        # unfortunately binance and probably some others don't support watching orders for multiple symbols
        # so we need to watch orders for each symbol separately
        async def _watch_executions():
            exec = await self._exchange.watch_orders(instrument.symbol)
            _msg = f"\nexecs_{instrument.symbol} = [\n"
            for report in exec:
                _msg += "\t" + str(report) + ",\n"
                order, deals = self.trading_service.process_execution_report(instrument, report)
                # - send update to client
                channel.send((instrument, "order", order))
                if deals:
                    channel.send((instrument, "deals", deals))
            # logger.debug(_msg + "]\n")

        await self._listen_to_stream(
            subscriber=_watch_executions,
            exchange=self._exchange,
            channel=channel,
            name=name,
        )



from .utils import (
    srows,
    scols,
    continuous_periods,
    ohlc_resample,
    retain_columns_and_join,
    select_column_and_join,
    dict_to_frame,
    drop_duplicated_indexes,
    process_duplicated_indexes,
    generate_equal_date_ranges,
    rolling_forward_test_split,
    shift_series,
    check_frame_columns,
)



from .readers import (
    DataReader,
    CsvStorageDataReader,
    MultiQdbConnector,
    QuestDBConnector,
    AsOhlcvSeries,
    AsPandasFrame,
    AsQuotes,
    AsTimestampedRecords,
    RestoreTicksFromOHLC,
)

from .helpers import loader



import traceback

from typing import Any, Callable, List, Dict
from threading import Thread

from qubx import logger
from qubx.core.account import AccountProcessor
from qubx.core.helpers import BasicScheduler, CachedMarketDataHolder, set_parameters_to_object
from qubx.core.loggers import StrategyLogging
from qubx.core.basics import Instrument, dt_64, SW, CtrlChannel
from qubx.core.loggers import StrategyLogging
from qubx.data.readers import DataReader
from qubx.gathering.simplest import SimplePositionGatherer
from qubx.trackers.sizers import FixedSizer
from qubx.core.interfaces import (
    IBrokerServiceProvider,
    IMarketDataProvider,
    IPositionGathering,
    IStrategy,
    ITradingServiceProvider,
    IStrategyContext,
    IUniverseManager,
    ISubscriptionManager,
    ITradingManager,
    IProcessingManager,
    PositionsTracker,
)
from .mixins import ProcessingManager, SubscriptionManager, TradingManager, UniverseManager, MarketDataProvider


class StrategyContext(IStrategyContext):
    DEFAULT_POSITION_TRACKER: Callable[[], PositionsTracker] = lambda: PositionsTracker(
        FixedSizer(1.0, amount_in_quote=False)
    )

    __market_data_provider: IMarketDataProvider
    __universe_manager: IUniverseManager
    __subscription_manager: ISubscriptionManager
    __trading_manager: ITradingManager
    __processing_manager: IProcessingManager

    __trading_service: ITradingServiceProvider  # service for exchange API: orders managemewnt
    __logging: StrategyLogging  # recording all activities for the strat: execs, positions, portfolio
    __broker: IBrokerServiceProvider  # market data provider
    __cache: CachedMarketDataHolder
    __scheduler: BasicScheduler
    __initial_instruments: list[Instrument]

    __thread_data_loop: Thread | None = None  # market data loop
    __is_initialized: bool = False

    def __init__(
        self,
        strategy: IStrategy,
        broker: IBrokerServiceProvider,
        account: AccountProcessor,
        instruments: list[Instrument],
        logging: StrategyLogging,
        config: dict[str, Any] | None = None,
        position_gathering: IPositionGathering | None = None,  # TODO: make position gathering part of the strategy
        aux_data_provider: DataReader | None = None,
    ) -> None:
        self.account = account
        self.strategy = self.__instantiate_strategy(strategy, config)

        self.__broker = broker
        self.__logging = logging
        self.__scheduler = broker.get_scheduler()
        self.__trading_service = broker.get_trading_service()
        self.__trading_service.set_account(self.account)
        self.__initial_instruments = instruments

        self.__cache = CachedMarketDataHolder()

        __position_tracker = self.strategy.tracker(self)
        if __position_tracker is None:
            __position_tracker = StrategyContext.DEFAULT_POSITION_TRACKER()

        __position_gathering = position_gathering if position_gathering is not None else SimplePositionGatherer()

        self.__market_data_provider = MarketDataProvider(
            cache=self.__cache,
            broker=self.__broker,
            universe_manager=self,
            aux_data_provider=aux_data_provider,
        )
        self.__universe_manager = UniverseManager(
            context=self,
            strategy=self.strategy,
            broker=self.__broker,
            trading_service=self.__trading_service,
            cache=self.__cache,
            logging=self.__logging,
            subscription_manager=self,
            trading_manager=self,
            time_provider=self,
            account_processor=self.account,
            position_gathering=__position_gathering,
        )
        self.__subscription_manager = SubscriptionManager(broker=self.__broker)
        self.__trading_manager = TradingManager(
            time_provider=self,
            trading_service=self.__trading_service,
            strategy_name=self.strategy.__class__.__name__,
        )
        self.__processing_manager = ProcessingManager(
            context=self,
            strategy=self.strategy,
            broker=self.__broker,
            logging=self.__logging,
            market_data=self,
            subscription_manager=self,
            time_provider=self,
            position_tracker=__position_tracker,
            position_gathering=__position_gathering,
            cache=self.__cache,
            scheduler=self.__broker.get_scheduler(),
            is_simulation=self.__broker.is_simulated_trading,
        )
        self.__post_init__()

    def __post_init__(self) -> None:
        self.strategy.on_init(self)
        # - update cache default timeframe
        _, __sub_params = self.get_base_subscription()
        __default_timeframe = __sub_params.get("timeframe", "1Sec")
        self.__cache.update_default_timeframe(__default_timeframe)

    def time(self) -> dt_64:
        return self.__trading_service.time()

    def start(self, blocking: bool = False):
        if self.__is_initialized:
            raise ValueError("Strategy is already started !")

        # - run cron scheduler
        self.__scheduler.run()

        # - create incoming market data processing
        databus = self.__broker.get_communication_channel()
        databus.register(self)

        # - update universe with initial instruments after the strategy is initialized
        self.set_universe(self.__initial_instruments, skip_callback=True)

        # - initialize strategy (should we do that after any first market data received ?)
        if not self.__is_initialized:
            try:
                self.strategy.on_start(self)
                self.__is_initialized = True
            except Exception as strat_error:
                logger.error(
                    f"(StrategyContext) Strategy {self.strategy.__class__.__name__} raised an exception in on_start: {strat_error}"
                )
                logger.error(traceback.format_exc())
                return

        # - for live we run loop
        if not self.__broker.is_simulated_trading:
            self.__thread_data_loop = Thread(target=self.__process_incoming_data_loop, args=(databus,), daemon=True)
            self.__thread_data_loop.start()
            logger.info("(StrategyContext) strategy is started in thread")
            if blocking:
                self.__thread_data_loop.join()

    def stop(self):
        if self.__thread_data_loop:
            self.__broker.close()
            self.__broker.get_communication_channel().stop()
            self.__thread_data_loop.join()
            try:
                self.strategy.on_stop(self)
            except Exception as strat_error:
                logger.error(
                    f"(StrategyContext) Strategy {self.strategy.__class__.__name__} raised an exception in on_stop: {strat_error}"
                )
                logger.opt(colors=False).error(traceback.format_exc())
            self.__thread_data_loop = None

        # - close logging
        self.__logging.close()

    @property
    def is_simulation(self) -> bool:
        return self.__broker.is_simulated_trading

    # IAccountViewer delegation
    @property
    def positions(self):
        return self.account.positions

    def get_capital(self) -> float:
        return self.account.get_capital()

    def get_total_capital(self) -> float:
        return self.account.get_total_capital()

    def get_reserved(self, instrument: Instrument) -> float:
        return self.account.get_reserved(instrument)

    # IMarketDataProvider delegation
    def ohlc(self, instrument: Instrument, timeframe: str | None = None, length: int | None = None):
        return self.__market_data_provider.ohlc(instrument, timeframe, length)

    def quote(self, instrument: Instrument):
        return self.__market_data_provider.quote(instrument)

    def get_data(self, instrument: Instrument, sub_type: str) -> List[Any]:
        return self.__market_data_provider.get_data(instrument, sub_type)

    def get_aux_data(self, data_id: str, **parameters):
        return self.__market_data_provider.get_aux_data(data_id, **parameters)

    def get_instruments(self):
        return self.__market_data_provider.get_instruments()

    def get_instrument(self, symbol: str, exchange: str):
        return self.__market_data_provider.get_instrument(symbol, exchange)

    # ITradingManager delegation
    def trade(self, instrument: Instrument, amount: float, price: float | None = None, time_in_force="gtc", **options):
        return self.__trading_manager.trade(instrument, amount, price, time_in_force, **options)

    def cancel(self, instrument: Instrument):
        return self.__trading_manager.cancel(instrument)

    def cancel_order(self, order_id: str):
        return self.__trading_manager.cancel_order(order_id)

    # IUniverseManager delegation
    def set_universe(self, instruments: list[Instrument], skip_callback: bool = False):
        return self.__universe_manager.set_universe(instruments, skip_callback)

    @property
    def instruments(self):
        return self.__universe_manager.instruments

    # ISubscriptionManager delegation
    def subscribe(self, instruments: List[Instrument] | Instrument, subscription_type: str | None = None, **kwargs):
        return self.__subscription_manager.subscribe(instruments, subscription_type, **kwargs)

    def unsubscribe(self, instruments: List[Instrument] | Instrument, subscription_type: str | None = None):
        return self.__subscription_manager.unsubscribe(instruments, subscription_type)

    def has_subscription(self, instrument: Instrument, subscription_type: str):
        return self.__subscription_manager.has_subscription(instrument, subscription_type)

    def get_subscriptions(self, instrument: Instrument) -> Dict[str, Dict[str, Any]]:
        return self.__subscription_manager.get_subscriptions(instrument)

    def get_base_subscription(self):
        return self.__subscription_manager.get_base_subscription()

    def set_base_subscription(self, subscription_type, **kwargs):
        return self.__subscription_manager.set_base_subscription(subscription_type, **kwargs)

    def get_warmup(self, subscription_type: str):
        return self.__subscription_manager.get_warmup(subscription_type)

    def set_warmup(self, subscription_type: str, period: str):
        return self.__subscription_manager.set_warmup(subscription_type, period)

    # IProcessingManager delegation
    def process_data(self, instrument: Instrument, d_type: str, data: Any):
        return self.__processing_manager.process_data(instrument, d_type, data)

    def set_fit_schedule(self, schedule: str):
        return self.__processing_manager.set_fit_schedule(schedule)

    def set_event_schedule(self, schedule: str):
        return self.__processing_manager.set_event_schedule(schedule)

    # private methods
    def __process_incoming_data_loop(self, channel: CtrlChannel):
        logger.info("(StrategyContext) Start processing market data")
        while channel.control.is_set():
            with SW("StrategyContext._process_incoming_data"):
                # - waiting for incoming market data
                instrument, d_type, data = channel.receive()
                if self.process_data(instrument, d_type, data):
                    channel.stop()
                    break
        logger.info("(StrategyContext) Market data processing stopped")

    def __instantiate_strategy(self, strategy: IStrategy, config: dict[str, Any] | None) -> IStrategy:
        __strategy = strategy() if isinstance(strategy, type) else strategy
        __strategy.ctx = self
        set_parameters_to_object(__strategy, **config if config else {})
        return __strategy



from .market import MarketDataProvider
from .processing import ProcessingManager
from .subscription import SubscriptionManager
from .trading import TradingManager
from .universe import UniverseManager
